from collections.abc import Mapping
import datetime
from typing import Optional, Union

from sqlalchemy import orm
from OpenSSL import crypto

from .configuration_object import ConfigurationObject
from .domain import Domain
from globus.manager.constants import ConfigurationObjectTag
from globus.manager.api.schemas.appsync.certificate import certificate_schema
from globus.manager.api.schemas.api.domain import DomainDict
from globus.manager.utils.run_as_user import run_as_user


class LetsEncryptCertificate(ConfigurationObject):
    """
    Model for a managed Let's Encrypt certificate for a wildcard domain. This
    uses the Domain model to handle writing the actual certificate and key to
    the filesystem so that apache/gridftp can find it.
    """

    _DATA_TYPE = "certificate#1.1.0"
    __mapper_args__ = {"polymorphic_identity": ConfigurationObjectTag.CERTIFICATE.name}
    PROPERTIES = {
        "certificate#1.0.0": ["domain_name", "certificate", "key", "chain"],
        "certificate#1.1.0": ["domain_name", "certificate", "key", "chain", "domain"],
    }
    configuration_schema = certificate_schema

    def __init__(self, appsync_data=None, doc=None):
        self._x509_certificate = None
        self._domain = None  # type: Optional[Domain]

        if appsync_data is not None:
            args = {k: v for k, v in appsync_data.items()}
            args["tag"] = ConfigurationObjectTag.CERTIFICATE
            args["id"] = ConfigurationObjectTag.CERTIFICATE.name
            super().__init__(**args)
            self.init_on_load()
        elif doc is not None and (doc.get("DATA_TYPE") == self._DATA_TYPE):
            super().__init__(
                tag=ConfigurationObjectTag.CERTIFICATE,
                id=ConfigurationObjectTag.CERTIFICATE.name,
            )
            self._put(doc)
            self.refresh_configuration_hash()
        else:
            raise ValueError("Unsupported DATA_TYPE")

    @orm.reconstructor
    def init_on_load(self):
        self._x509_certificate = None
        self._domain = None
        self._put(self.configuration_hash)

    def _put(self, doc):
        for k, v in doc.items():
            if k in self.PROPERTIES[self._DATA_TYPE]:
                if k != "domain":
                    setattr(self, k, v)

        if not self._domain:
            self.domain = Domain(
                doc={
                    "DATA_TYPE": "domain#1.0.0",
                    "wildcard": True,
                    "domain_name": self.domain_name,
                    "certificate": self.certificate,
                    "certificate_chain": self.chain,
                    "private_key": self.key,
                }
            )

    domain_name = ConfigurationObject.basic_property("domain_name")
    certificate = ConfigurationObject.basic_property("certificate")
    key = ConfigurationObject.basic_property("key")
    chain = ConfigurationObject.basic_property("chain")

    @property
    def domain(self) -> Domain:
        """
        Details about the certifiate's domain, including paths to the
        certificates and keys on the data transfer node.
        """
        if self._domain is not None:
            return self._domain
        else:
            raise AttributeError("The domain property is not defined for this object")

    @domain.setter
    def domain(self, value: Optional[Union[DomainDict, Domain]]) -> None:
        if isinstance(value, Domain):
            self._domain = value
        elif isinstance(value, Mapping):
            self._domain = Domain(appsync_data=value)
        else:
            self._domain = None

    @property
    def expiration_time(self):
        """
        Expiration time as an zone-aware datetime object
        """
        cert = self.x509_certificate
        ts = cert.get_notAfter().decode()

        if ts.endswith("Z"):
            ts = ts[:-1] + "+0000"

        return datetime.datetime.strptime(ts, "%Y%m%d%H%M%S%z")

    @property
    def x509_certificate(self):
        """
        OpenSSL X.509 representation of this certificate
        """
        if not self._x509_certificate:
            self._x509_certificate = crypto.load_certificate(
                crypto.FILETYPE_PEM, self.certificate
            )
        return self._x509_certificate

    @staticmethod
    def create_certificate(domain):
        acme_account = ConfigurationObject.query.filter(
            ConfigurationObject.tag == ConfigurationObjectTag.ACME_ACCOUNT
        ).one()

        return LetsEncryptCertificate(
            doc=dict(
                DATA_TYPE=LetsEncryptCertificate._DATA_TYPE,
                **acme_account.get_cert_chain_and_key(domain)._asdict()
            )
        )

    def renew(self):
        acme_account = ConfigurationObject.query.filter(
            ConfigurationObject.tag == ConfigurationObjectTag.ACME_ACCOUNT
        ).one()

        res = acme_account.get_cert_chain_and_key(self.domain_name)
        # Record the result in our attributes
        self.certificate = res.certificate
        self.key = res.key
        self.chain = res.chain
        # Also record the result in our domain
        self.domain.certificate = res.certificate
        self.domain.private_key = res.key
        self.domain.certificate_chain = res.chain
        self.configuration_hash.changed()

    @property
    def files_exist(self):
        return self.domain.files_exist

    @property
    def certificate_path(self):
        return self.domain.certificate_path

    @property
    def certificate_chain_path(self):
        return self.domain.certificate_chain_path

    @property
    def private_key_path(self):
        return self.domain.private_key_path

    @run_as_user(user="gcsweb")
    def write_config(self, delete=False):
        return self.domain.write_config(delete=delete)

    def __repr__(self):
        return "<LetsEncryptCertificate {}>".format(self.domain_name)
